/* BitTorrentfs - a BitTorrent client file server */
#include <u.h>
#include <libc.h>
#include <bio.h>
#include <fcall.h>
#include <thread.h> 
#include <9p.h>
#include <pool.h>
#include "torrent.h"
#include "torrentfile.h"
#include "misc.h"
#include "comm.h"

Torrent *torrents[10];
char mypeerid[PEERIDLEN+1];
char port[] = "6881";
char *datadir;
int verbose = 0;
extern Srv fs;

static void
setpeerid(void)
{
	/*
	peer id = 20 bytes = <"-p9btfs-"><12 random chars>
	*/
	char prefix[] = "-p9btfs-";
	//char prefix[] = "M3-4-2--";
	long prefixsz = strlen(prefix);
	int suffixsz = PEERIDLEN - prefixsz;

	strncpy(mypeerid, prefix, prefixsz);
	for (int i=0;i<suffixsz;i++){
		mypeerid[i+prefixsz] = (char)pickrand(65,90);
	}
	mypeerid[PEERIDLEN] = '\0';
}

static void
setdatadir(char *dir)
{
	uchar length = strlen(dir);
	datadir = emalloc(length+1);
	datadir = strcpy(datadir,dir);
}

char *
propername(Rune *name, uchar maxsize)
{
	char *buf;
	uchar namelen;

	/*
	pass 0 as a maxsize for no maxsize
	*/
	if ((namelen = runestrlen(name)) > maxsize && maxsize != 0)
		namelen = maxsize;
	buf = emalloc(namelen+1);
	for (int i=0; i<namelen; i++){
		runetochar(&(buf[i]),&(name[i]));	
		if (buf[i] == ' ')
			buf[i] = '_';
	}
	buf[namelen] = '\0';
	return buf;
}

static void
tortotree(File *root, Torrent *tor)
{
	File *torrent;
	char *buf;
	File *parent;
	File *tracker;
	File *announce;
	File *files;

	buf = emalloc(2 * HASHSIZE + 1);
	for (int i=0; i<HASHSIZE; i++)
		sprint(&(buf[2*i]),"%.2ux", tor->infohash[i]);
	parent = root;
	incref(parent);
	parent = walkfile(parent, "torrents");
	if(parent == nil){
		free(buf);
		closefile(parent);
		return;
	}
	torrent = createfile(parent, buf, parent->uid, DMDIR|0644, nil);
	tracker = createfile(torrent, "tracker", torrent->uid, DMDIR|0644, nil);
	announce = createfile(tracker, "announce", tracker->uid, 0644, nil);
	files = createfile(torrent, "pieces", torrent->uid, DMDIR|0644, nil);
	/*
	files->aux = smprint("%s", "pieces");
	print("%s \n", files->aux);
	*/

	closefile(parent);
	free(buf);
}

static void 
filltree(File *root)
{
	File *ctl;
	File *torrents;

	ctl = createfile(root, "ctl", root->uid, 0644, nil);
	torrents = createfile(root, "torrents", root->uid, DMDIR|0644, nil);
}

Torrent *
addtorrent(char *torrentfile)
{
	Torrent *tor = nil;
	Piece *lister;

	print("Adding %s\n", torrentfile);
	tor = emalloc(sizeof(Torrent));
	parsebtfile(torrentfile, tor);
	tor->infohash = getinfohash(torrentfile, tor->infosize);
	for (int i = 0; i<20; i++)
		print("%%%.2ux", tor->infohash[i]);
	print("\n");
	tortotree(fs.tree->root, tor);

	/* create data files */
	tor->datafiles = emalloc(tor->filesnb * sizeof(char *));
	for (int i=0; i<tor->filesnb; i++){	
		tor->datafiles[i] = emalloc(strlen(datadir) + 1 + strlen(propername(tor->filepath[i], 0))+ 1);
		tor->datafiles[i] = strcpy(tor->datafiles[i], datadir);
		tor->datafiles[i] = strcat(tor->datafiles[i], "/");
		tor->datafiles[i] = strcat(tor->datafiles[i], propername(tor->filepath[i], 0));
		dbgprint("tor->datafiles[%d]: %s\n", i, tor->datafiles[i]);
		createpath(tor->datafiles[i]);
	}
	preppieceslist(tor);
	/*
	lister = tor->pieceslist;
	while (lister != nil) {
		dbgprint("[%d, %d] ",lister->index, lister->status);
		lister = lister->next;	
	}
	dbgprint("\n");
	*/
	calltracker(tor, "announce");
	return tor;
}

void
leecher(void *arg)
{
	struct Params{ Torrent *tor; Peer *peer; Channel *c;} *params;

	params = arg;
	print("starting leeching\n");
	callpeer(params->peer,params->tor,params->c);
	print("done leeching\n");
	free(params);
	threadexits(0); 
}

void
leechers(void *arg)
{
	Torrent *tor = arg;
	int i=0;
	char m[1];
	int n;
	Alt *a;
	struct Params{ Torrent *tor; Peer *peer; Channel *c;} *params;

	while (i<tor->peersnb){
		a = erealloc(a, (i+1)*sizeof(Alt));
		a[i].v = m;
		a[i].c = chancreate(sizeof(m), 0);
		a[i].op = CHANRCV;
		params = emalloc(sizeof(struct Params));
		params->tor = tor;
		params->peer = tor->peers[i];
		params->c = a[i].c;
		tor->peers[i]->busy = 1;
		threadcreate(leecher, params, STACK);
		dbgprint("thread for peer #%d started\n", i);
		i++;
	}
	a = erealloc(a, (i+1)*sizeof(Alt));
	a[i].v = nil;
	a[i].c = nil;
	a[i].op = CHANEND;
	for(;;){
		n = alt(a);
		if((n<0) || (n>i))
			error("with alt");
		print("thread #%d taking over\n",n);
		sleep(1000);
	}
	threadexits(0);	
}

static void
usage(void)
{
	fprint(2, "usage: btfs [-d datadir] [-m mntpt] ");
	exits("usage");
}

void
threadmain(int argc, char **argv)
{
	mainmem->flags |= POOL_PARANOIA;
	char *mtpt = "/n/btfs";
	char *datadir = getenv("home");
	//char *datadir = "/usr/glenda/torrents";
	Dir *dir;
	int fd;
	
	ARGBEGIN{
	case 'm':
		mtpt = ARGF();
		break;
	case 'd':
		datadir = ARGF();
		break;
	case 'v':
		verbose = 1;
	default:
		break;
	}ARGEND

	if ((dir = dirstat(mtpt)) == nil) {
		print("%s does not exist, creating it. \n", mtpt);
		fd = create(mtpt,OREAD,DMDIR|0755);		
		close(fd);
	}
	free(dir);

	setdatadir(datadir);
	if ((dir = dirstat(datadir)) == nil) {
		print("%s does not exist, creating it. \n", datadir);
		fd = create(datadir,OREAD,DMDIR|0755);	
		close(fd);	
	}
	free(dir);

	fs.tree = alloctree(nil, nil, DMDIR|0777, nil);
	filltree(fs.tree->root);
	threadpostmountsrv(&fs, nil, mtpt, MREPL|MCREATE);
	setpeerid();
	//torrents[0] = addtorrent("/usr/glenda/ebooks.torrent");
	//leechers(torrents[0]);
	threadexits(0);
}
