#include <u.h>
#include <libc.h>
#include <libsec.h>
#include <pool.h>
#include "misc.h"
#include "torrent.h"

extern char mypeerid[];
extern char port[];

void 
preppieceslist(Torrent *tor)
{
	/*
	temp is used for operations with many uvlongs, because
	not enough registers to handle them in one go.
	*/
	uvlong fullpieces, temp;
	uint piecelength;
	int *offset;
	uvlong *length;
	Piece *lister;

	length = tor->filelength;
	piecelength = tor->piecelength;
	tor->pieceslist = emalloc(sizeof(Piece));
	lister = tor->pieceslist;
	for (int i=0; i<tor->piecesnb-1; i++){
		lister->index = i;
/*
		if(i<29 || i>71)
			lister->status = BThas;
		else
*/
			lister->status = 0;
		lister->next = emalloc(sizeof(Piece));
		lister = lister->next;
	}
	lister->index = tor->piecesnb-1;
	lister->status = 0;
	lister->next = 	nil;
	tor->firstoffset = emalloc(tor->filesnb * sizeof(int));
	offset = tor->firstoffset;
	offset[0] = 0;
	tor->fileborder = emalloc(tor->filesnb*sizeof(int));
	tor->fileborder[0] = 0;
	for (int i=0; i<tor->filesnb; i++){
		/* file is smaller than a piece */
		temp = offset[i] + length[i];
		if(temp < piecelength){
			if(offset[i] == 0 && i < tor->filesnb - 1){
				lister = tor->pieceslist;
				while ((lister != nil) && (lister->index != tor->fileborder[i]))
					lister = lister->next;
				lister->status |= BTshar;
			}
			if(i < tor->filesnb - 1){
				tor->fileborder[i+1] = tor->fileborder[i];
				temp = offset[i] + length[i];
				offset[i+1] = temp;
			} else
				tor->lastpiece = length[i];
		} else{
			if(offset[i] != 0){
				temp = length[i];
				temp -= piecelength - offset[i];
				fullpieces = temp / piecelength;
				temp = length[i];
				temp -= fullpieces * piecelength;
				if(i < tor->filesnb - 1){
					tor->fileborder[i+1] = tor->fileborder[i] + fullpieces + 1;
					offset[i+1] = temp - (piecelength - offset[i]);
				} else
					tor->lastpiece = temp - (piecelength - offset[i]);
			} else{
				fullpieces = length[i] / piecelength;
				temp = length[i];
				if(i < tor->filesnb - 1){
					tor->fileborder[i+1] = tor->fileborder[i] + fullpieces;
					offset[i+1] = temp - fullpieces * piecelength;
				} else
					tor->lastpiece = temp - fullpieces * piecelength;
			}
			if(i < tor->filesnb - 1 && offset[i+1] != 0){
				lister = tor->pieceslist;
				while ((lister != nil) && (lister->index != tor->fileborder[i+1]))
					lister = lister->next;
				lister->status |= BTshar;
			}
		}
	}
}

void 
preppeerspieces(Torrent *tor, Peer *peer)
{
	char *bitfield;
	Piece *lister;
	Piece *rimmer;
	Piece *kryten;
	int index;
	int byteindex;
	char bitindex;
	char mask;

	bitfield = peer->bitfield;
	lister = tor->pieceslist;
	peer->pieceslist = emalloc(sizeof(Piece));
	rimmer = peer->pieceslist;
	kryten = rimmer;
	peer->piecesnb = 0;
	while (lister != nil){
		index = lister->index;
		byteindex = index / 8;
		bitindex = index % 8 +1;
		mask = 1 << (8 - bitindex); 
		if ((bitfield[byteindex] & mask) != 0){
			rimmer->index = index;
			rimmer->status = lister->status;
			rimmer->next = emalloc(sizeof(Piece));
			kryten = rimmer;
			rimmer = rimmer->next;
			peer->piecesnb++;
		}
		lister = lister->next;
	}
	free(rimmer);
	kryten->next = nil;
}

char *
forgerequest(Torrent *tor, char *req)
{
	char *announce;
	int length;
	char *baseurl;
	char infohash[3*HASHSIZE+1]; 
	char peerid[3*PEERIDLEN+1]; 
	int len = strlen(tor->announce);
	char *buf = emalloc((512)*sizeof(char));

	/*
	Is it really fair to bail out if announce url
	doesn't contain "announce"?
	*/
	if((announce = strstr(tor->announce, "announce")) == 0)
		error("weird announce url");
	length = strlen(announce);
	len = len - length;

	baseurl = emalloc((len+1)*sizeof(char));
	baseurl = strncpy(baseurl,tor->announce,len);
	baseurl[len] = '\0';		

	for (int i = 0; i<HASHSIZE; i++){
		infohash[3*i] = '%';
		sprint(&(infohash[3*i+1]),"%.2ux", tor->infohash[i]);
	}
	infohash[3*HASHSIZE] = '\0';

	for (int i = 0; i<20; i++){
		peerid[3*i] = '%';
		sprint(&(peerid[3*i+1]),"%.2x", mypeerid[i]);
	}
	infohash[3*PEERIDLEN] = '\0';

	if (strcmp(req, "scrape") == 0){
		/*
		No need to realloc precisely, just use a big enough buffer
		redo it like for announce

		buf = erealloc(buf,(len+7+10+3*HASHSIZE+1)*sizeof(char));
		buf = strcpy(buf, baseurl);
		buf = strcat(buf,"scrape?info_hash=");
		buf = strcat(buf,infohash);
		*/
	}
	else if (strcmp(req, "announce") == 0){
		buf = strcpy(buf, baseurl);
		buf = strcat(buf, announce);
		buf = strcat(buf,"?info_hash=");
		buf = strcat(buf,infohash);
		buf = strcat(buf,"&port=");
		buf = strcat(buf,port);
		buf = strcat(buf,"&uploaded=0&downloaded=0&left=0&event=started");
/*
sending compact=0 will in fact result in compact replies! 
*/
		//buf = strcat(buf,"&compact=0");
		//buf = strcat(buf,"&ip=127.0.0.1");
		buf = strcat(buf,"&peer_id=");
		buf = strcat(buf,peerid);
	}

	free(baseurl);
	return buf;
}

void
writedata(ulong index, char *data, Torrent *tor, Peer *peer)
{
	int m, fd, fileindex, towrite, offset, piecelength;
	uvlong *filelength;
	Piece *lister;
	Piece *remover;

	fileindex = 0;
	for (int i=tor->filesnb-1; i>=0; i--)
		if (index > tor->fileborder[i]){
			fileindex = i;
			break;
		}
	piecelength = tor->piecelength;
	filelength = tor->filelength;
	/* case when file is smaller than piece is still missing */
	if (tor->firstoffset[fileindex] != 0)
		offset = piecelength - tor->firstoffset[fileindex] + (index - tor->fileborder[fileindex] - 1) * piecelength;
	else
		offset = (index - tor->fileborder[fileindex]) * piecelength;
	if ((fd = open(tor->datafiles[fileindex], OWRITE)) < 0)
		sysfatal("couldn't open %s: %r", tor->datafiles[fileindex]);

	dbgprint("offset in file: %d\n", offset);
	lister = tor->pieceslist;
	while ((lister != nil) && (lister->index != index))
			lister = lister->next;	
	if (((lister->status)&BTshar) != 0){
		towrite = tor->firstoffset[fileindex+1];
		if ((m = pwrite(fd, data, towrite, offset)) != m)
				sysfatal("could not write partial piece to disk: %r");
		close(fd);
		dbgprint("shared piece: %ld\n", index);
		dbgprint("wrote %d bytes at offset %d \n", towrite, offset);
		while(1){
			sleep(2000);
			fileindex++;
			if ((fd = open(tor->datafiles[fileindex], OWRITE)) < 0)
				sysfatal("couldn't open %s: %r", tor->datafiles[fileindex]);	
			offset = 0;
			if ((index == tor->piecesnb-1) && (fileindex == tor->filesnb-1)){
				towrite = tor->lastpiece;
				if ((m = pwrite(fd, &data[tor->firstoffset[fileindex]], towrite, offset)) != m)
					sysfatal("could not write partial piece to disk: %r");
				dbgprint("then wrote %d bytes at %d\n", towrite, offset);
				break;
			} else{
				if(tor->fileborder[fileindex+1] == index){
					towrite = filelength[fileindex];
					if ((m = pwrite(fd, &data[tor->firstoffset[fileindex]], towrite, offset)) != m)
						sysfatal("could not write partial piece to disk: %r");
					dbgprint("then wrote %d bytes at %d\n", towrite, offset);
					close(fd);
				} else{
					towrite = piecelength - tor->firstoffset[fileindex];
					if ((m = pwrite(fd, &data[tor->firstoffset[fileindex]], towrite, offset)) != m)
						sysfatal("could not write partial piece to disk: %r");
					dbgprint("then wrote %d bytes at %d\n", towrite, offset);
					break;
				}
			}
		}
	} else{
		if (index == tor->piecesnb-1){
			if ((m = pwrite(fd, data, tor->lastpiece, offset)) != m)
				sysfatal("could not write last piece to disk: %r");
		}
		else{
			towrite = piecelength;
			if ((m = pwrite(fd, data, towrite, offset)) != m)
				sysfatal("could not write piece to disk: %r");
		}
	}
	close(fd);
	/* update the list */
	lister = tor->pieceslist;
	while ((lister != nil) && (lister->index != index))
		lister = lister->next;
	lister->status |= BThas;
	/* now for the peer's list */
	lister = peer->pieceslist;
	remover = lister;
	while ((lister != nil) && (lister->index != index)){
		remover = lister;
		lister = lister->next;
	}
	if (lister == peer->pieceslist)
		peer->pieceslist = lister->next;
	remover->next = lister->next;
	free(lister);
	peer->piecesnb--;
	dbgprint("wrote piece: %ld\n", index);
}

int
checkpiece(char *data, Torrent *tor, int index)
{
	uchar *digest;
	uchar *buf;
	int datasize;
	Piece *lister;

	if (index == tor->piecesnb-1){
		lister = tor->pieceslist;
		while ((lister != nil) && (lister->index != index))
		lister = lister->next;
		/*
		if piece is shared, lastpiece is actually not the
		size of the whole piece. Too lazy to figure out 
		the right algo right now, so assume it's all good.
		*/
		if ((lister->status & BTshar) != 0)
			return 1;
		else
			datasize = tor->lastpiece;
	} else
		datasize = tor->piecelength;
	digest = emalloc(SHA1dlen);
	buf = emalloc(datasize);

	memmove(buf, data, datasize);
	sha1(buf, datasize, digest, nil);

	for (int i=0; i<HASHSIZE; i++){
		//print("%%%.2ux",digest[i]);
		if (digest[i] != tor->sha1list[index][i]){
			free(digest);
			free(buf);
			return -1;	
		}
	}
	free(digest);
	free(buf);
	return 1;
}

